"""
Package to helper to encode class in json
"""

import json
import typing


class ToJSON:
    """
    ToJSON will provide a dict() and a to_json() methods for your class
    """

    @staticmethod
    def set_encoder_resolver(attr_type, resolver_callback: typing.Callable):
        """
        Register a resolver for a class or subclass
        attr_type: int,str,bool,float or any other class
        resolver_callback: lambda function that receives the value to be cast.
         Example 'lambda value: value'
        """
        JSONEncoder.base_types_resolver.update({attr_type: resolver_callback})

    def dict(self):
        """
        Returns dict representation of your object
        """
        _dict_representation = {}
        _annotations = typing.get_type_hints(type(self))

        del _annotations["_Kobject__custom_exception"]

        if "_FromJSON__custom_exception" in _annotations:
            del _annotations["_FromJSON__custom_exception"]

        for attr in _annotations:
            attr_value = object.__getattribute__(self, attr)
            _dict_representation.update({attr: attr_value})
        return _dict_representation

    def to_json(self) -> bytes:
        """
        Returns json of your object
        """
        dict_repr = self.dict()
        json_bytes = json.dumps(dict_repr, default=JSONEncoder.default)
        return json_bytes.encode()


class JSONEncoder(json.JSONEncoder):
    """
    Self implementation for default json.JSONEncoder
    """

    base_types_resolver = {}

    @classmethod
    def default(cls, obj):  # pylint: disable=W0221
        if isinstance(obj, ToJSON):
            return obj.dict()
        for map_attr_type, resolver in cls.base_types_resolver.items():
            if isinstance(obj, map_attr_type):
                return resolver(obj)
        return obj



"""
Package to  helper to decode json in a class instance
"""

import inspect
import json
import typing


T = typing.TypeVar("T")


class FromJSON:
    """
    FromJSON will provide a from_json() and a from_dict() methods to instantiate your objects
    """

    __custom_exception: typing.Type[Exception] = None

    @classmethod
    def set_custom_exception(cls, exception: typing.Type[Exception]):
        """
        Will change de default JSONDecodeError error
        """
        cls.__custom_exception = exception

    @staticmethod
    def set_decoder_resolver(attr_type, resolver_callback: typing.Callable):
        """
        Register a resolver for a class or subclass\n

        attr_type: int,str,bool,float or any other class\n
        resolver_callback: lambda function that receives the type and value to be cast.
         Example 'lambda attr_type, value: value'
        """
        JSONDecoder.types_resolver.insert(0, (attr_type, resolver_callback))

    @classmethod
    def from_json(cls: T, payload: bytes) -> typing.Type[T]:
        """
        Returns a class instance by the giving json payload
        """
        try:
            dict_repr = json.loads(payload)
        except Exception as original_error:
            if cls.__custom_exception is not None:
                raise cls.__custom_exception(  # pylint: disable=E1102
                    f"Invalid content -> {str(payload)}",
                ) from original_error
            raise original_error
        instance = cls.from_dict(dict_repr=dict_repr)
        return instance

    @classmethod
    def __attribute_has_default_value(cls, attr: str, attr_value: any) -> bool:
        _default_values = inspect.signature(cls.__init__).parameters
        default_value = _default_values.get(attr).default
        return attr_value == default_value

    @classmethod
    def from_dict(cls: T, dict_repr: dict) -> typing.Type[T]:
        """
        Returns a class instance by the giving dict representation
        """

        _annotations = typing.get_type_hints(cls)

        del _annotations["_Kobject__custom_exception"]
        del _annotations["_FromJSON__custom_exception"]

        for attr, attr_type in _annotations.items():
            attr_value = dict_repr.get(attr)
            casted_value = JSONDecoder.type_caster(
                attr_type=attr_type, attr_value=attr_value
            )
            is_default_value = cls.__attribute_has_default_value(
                attr=attr, attr_value=attr_value
            )
            if casted_value is not None or is_default_value:
                dict_repr[attr] = casted_value
                continue

            attr_type, sub_type = JSONDecoder.get_type(attr_type=attr_type)

            attr_value_new = []
            for attr_value_item in attr_value:
                attr_value_new.append(
                    JSONDecoder.type_caster(
                        attr_type=sub_type, attr_value=attr_value_item
                    )
                )
            dict_repr[attr] = attr_type(attr_value_new)
        instance = cls(**dict_repr)
        return instance


class JSONDecoder:
    """
    Implementation tyo decode json in to class instance inspired by json.JSONEncoder
    """

    types_resolver = []

    @staticmethod
    def get_type(attr_type):
        """
        Returns the type or subtypes of giving attribute type
        """
        sub_type = attr_type.__args__[0]
        attr_type = attr_type.__origin__
        return attr_type, sub_type

    @classmethod
    def type_caster(cls, attr_type, attr_value):
        """
        Returns the result of cast attribute type for the attribute type
        """
        for map_attr_type, resolver in cls.types_resolver:
            if inspect.isclass(attr_type) and issubclass(attr_type, map_attr_type):
                return resolver(attr_type, attr_value)
        return None


FromJSON.set_decoder_resolver(bool, lambda attr_type, value: value)
FromJSON.set_decoder_resolver(float, lambda attr_type, value: value)
FromJSON.set_decoder_resolver(int, lambda attr_type, value: value)
FromJSON.set_decoder_resolver(str, lambda attr_type, value: value)
FromJSON.set_decoder_resolver(
    FromJSON,
    lambda attr_type, value: attr_type.from_dict(value)
    if isinstance(value, dict)
    else value,
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kobject', 'kobject.from_json', 'kobject.to_json', 'kobject.validator']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'kobject',
    'version': '0.3.2',
    'description': 'Know your object is a attribute type checker',
    'long_description': '```\n                       ▄▄          ▄▄                      \n▀████▀ ▀███▀          ▄██          ██                 ██   \n  ██   ▄█▀             ██                             ██   \n  ██ ▄█▀      ▄██▀██▄  ██▄████▄  ▀███  ▄▄█▀██ ▄██▀████████ \n  █████▄     ██▀   ▀██ ██    ▀██   ██ ▄█▀   ███▀  ██  ██   \n  ██  ███    ██     ██ ██     ██   ██ ██▀▀▀▀▀▀█       ██   \n  ██   ▀██▄  ██▄   ▄██ ██▄   ▄██   ██ ██▄    ▄█▄    ▄ ██   \n▄████▄   ███▄ ▀█████▀  █▀█████▀    ██  ▀█████▀█████▀  ▀████\n                                ██ ██                      \n                                ▀███                       By CenturyBoys\n                                \nKnow your object is a __init__ type validator for class and dataclass\n```\n\n## Usage\n\nKobject can be use inside default class declaration and with dataclasses. Kobject uses the ```__init__``` signature to check types.\n\n### Default classes\n\n```python\nfrom kobject import Kobject\n\nclass StubClass(Kobject):\n    a_int: int\n    a_bool: bool\n    \n    def __init__(\n        self,\n        a_int: int,\n        a_bool: bool\n    ):\n        self.a_int = a_int\n        self.a_bool = a_bool\n        self.__post_init__()\n\ninstance = StubClass(a_int=1, a_bool=True)\n```\nNotice that in the default class declaration you need to call ```self.__post_init__()``` at the end of the ```__init__``` declaration.\n\n\n### Dataclass\n\n```python\nfrom dataclasses import dataclass\nfrom kobject import Kobject\n\n@dataclass\nclass StubClass(Kobject):\n    a_int: int\n    a_bool: bool\n\ninstance = StubClass(a_int=1, a_bool=True)\n```\nBy default, dataclass calls ```self.__post_init__()``` at the end of the ```__init__``` declaration.\n\n\n### Exception\n\nKobject raises ```TypeError``` with all validation errors, that means it checks all your object\'s attributes before raising the ```TypeError```. Types like List and Tuple will have all their elements checked.\n\n```python\nfrom dataclasses import dataclass\nfrom kobject import Kobject\nfrom typing import List, Tuple\n\n@dataclass\nclass StubClass(Kobject):\n    a_list_int: List[int]\n    a_tuple_bool: Tuple[bool]\n\ninstance = StubClass(a_list_int=[1, "", 2, ""], a_tuple_bool=["", True])\n```\n```bash\nTraceback (most recent call last):\n  File "/snap/pycharm-community/312/plugins/python-ce/helpers/pydev/pydevconsole.py", line 364, in runcode\n    coro = func()\n  File "<input>", line 10, in <module>\n  File "<string>", line 5, in __init__\n  File "/home/marco/projects/kobject/kobject/__init__.py", line 67, in __post_init__\n    raise TypeError(message)\nTypeError: Validation Errors:\n    \'a_list_int\' : Wrong type! Expected (<class \'int\'>,) but giving <class \'str\'> on index 1\n    \'a_list_int\' : Wrong type! Expected (<class \'int\'>,) but giving <class \'str\'> on index 3\n    \'a_tuple_bool\' : Wrong type! Expected <class \'tuple\'> but giving <class \'list\'>\n    \'a_tuple_bool\' : Wrong type! Expected (<class \'bool\'>,) but giving <class \'str\'> on index 0\n```\n\n### Default value\n\nKobject supports default values and will check them before any validation, that means if you declare a ```a_bool: bool = None``` it will not raise an error.\n\n```python\nfrom dataclasses import dataclass\nfrom kobject import Kobject\n\nclass StubClass(Kobject):\n    a_bool: bool = None\n\n    def __init__(self, a_bool: bool = 2):\n        self.a_bool = a_bool\n        self.__post_init__()\n\n@dataclass\nclass StubDataClass(Kobject):\n    a_bool: bool = None\n```\n\n### Custom exception\n\nBy default Kobject raise a ```TypeError``` but you can override this exception using ```set_custom_exception```\n\n```python\nfrom dataclasses import dataclass\nfrom kobject import Kobject\n\nclass CustomException(Exception):\n    pass\n\n\nKobject.set_custom_exception(CustomException)\n\n\n@dataclass\nclass StubClass(Kobject):\n    a__int: int\n\ninstance = StubClass(a__int="")\n```\n```bash\nTraceback (most recent call last):\n  File "/snap/pycharm-community/312/plugins/python-ce/helpers/pydev/pydevconsole.py", line 364, in runcode\n    coro = func()\n  File "<input>", line 15, in <module>\n  File "<string>", line 4, in __init__\n  File "/home/marco/projects/kobject/kobject/__init__.py", line 79, in __post_init__\n    raise exception(message)\n__main__.CustomException: Validation Errors:\n    \'a__int\' : Wrong type! Expected <class \'int\'> but giving <class \'str\'>\n```\n\n### ToJSON\n\nKobject has his own implementation to parse class instance to a JSON representation. \n\n```python\nfrom dataclasses import dataclass\nfrom typing import List, Tuple\n\nfrom kobject import Kobject, ToJSON\n    \n@dataclass\nclass BaseC(Kobject, ToJSON):\n    a_int: int\n    a_str: str\n    a_list_of_int: List[int]\n    a_tuple_of_bool: Tuple[bool]\n    \ninstance = BaseC(\n    a_int=1,\n    a_str="lala",\n    a_list_of_int=[1, 2, 3],\n    a_tuple_of_bool=(True,)\n)\n\njson_bytes = instance.to_json()\n\nprint(json_bytes)\n```\n```bash\nb\'{"a_int": 1, "a_str": "lala", "a_list_of_int": [1, 2, 3], "a_tuple_of_bool": [true]}\'\n```\n\nFor complex values ToJSON expose ```set_encoder_resolver``` to handler it.\n\n```python\nfrom dataclasses import dataclass\nfrom datetime import datetime\nfrom typing import List\nfrom uuid import UUID\n\nfrom kobject import Kobject, ToJSON, FromJSON\n\n\n@dataclass\nclass BaseA(Kobject, ToJSON):\n    a_datetime: datetime\n\n\n@dataclass\nclass BaseB:\n    a_uuid: UUID\n\n\n@dataclass\nclass BaseC(Kobject, ToJSON, FromJSON):\n    a_base_a: BaseA\n    a_base_b: BaseB\n    a_list_of_base_a: List[BaseA]\n\nToJSON.set_encoder_resolver(datetime, lambda value: str(value))\nToJSON.set_encoder_resolver(BaseB, lambda value: {"a_uuid": str(value.a_uuid)})\n\ninstance = BaseC(\n    a_base_a=BaseA(a_datetime=datetime.fromisoformat("2023-02-01 17:38:45.389426")),\n    a_base_b=BaseB(a_uuid=UUID("1d9cf695-c917-49ce-854b-4063f0cda2e7")),\n    a_list_of_base_a=[BaseA(a_datetime=datetime.fromisoformat("2023-02-01 17:38:45.389426"))]\n)\n\njson_bytes = instance.to_json()\n\nprint(json_bytes)\n```\n```bash\nb\'{"a_base_a": {"a_datetime": "2023-02-01 17:38:45.389426"}, "a_base_b": {"a_uuid": "1d9cf695-c917-49ce-854b-4063f0cda2e7"}, "a_list_of_base_a": [{"a_datetime": "2023-02-01 17:38:45.389426"}]}\'\n```\n\n### FromJSON\n\nKobject has his own implementation to parse JSON to a class instance. \n\n```python\nfrom dataclasses import dataclass\nfrom typing import List, Tuple\n\nfrom kobject import Kobject\nfrom kobject.from_json import FromJSON\n\n\n@dataclass\nclass BaseC(Kobject, FromJSON):\n    a_int: int\n    a_str: str\n    a_list_of_int: List[int]\n    a_tuple_of_bool: Tuple[bool]\n\npayload = (\n    b\'{"a_int": 1,"a_str": "lala","a_list_of_int": [1,2,3],\'\n    b\'"a_tuple_of_bool": [true]}\'\n)\ninstance = BaseC.from_json(payload=payload)\n\nprint(instance)\n```\n```bash\nBaseC(a_int=1, a_str=\'lala\', a_list_of_int=[1, 2, 3], a_tuple_of_bool=(True,))\n```\n\nFor complex values FromJSON expose ```set_decoder_resolver``` to handler it.\n\n```python\nfrom datetime import datetime\nfrom dataclasses import dataclass\nfrom typing import List\nfrom uuid import UUID\n\nfrom kobject import Kobject\nfrom kobject.from_json import FromJSON\n\n\n@dataclass\nclass BaseA(Kobject, FromJSON):\n    a_datetime: datetime\n\n\n@dataclass\nclass BaseB:\n    a_uuid: UUID\n\n\n@dataclass\nclass BaseC(Kobject, FromJSON):\n    a_base_a: BaseA\n    a_base_b: BaseB\n    a_list_of_base_a: List[BaseA]\n\nFromJSON.set_decoder_resolver(\n    datetime,\n    lambda attr_type, value: datetime.fromisoformat(value)\n    if isinstance(value, str)\n    else value,\n)\nFromJSON.set_decoder_resolver(\n    BaseB,\n    lambda attr_type, value: attr_type(a_uuid=UUID(value["a_uuid"]))\n    if isinstance(value, dict)\n    else value,\n)\npayload = (\n    b\'{"a_base_a": {"a_datetime": "2023-02-01 17:38:45.389426"},"a_base_b": {"a_\'\n    b\'uuid":"1d9cf695-c917-49ce-854b-4063f0cda2e7"}, "a_lis\'\n    b\'t_of_base_a": [{"a_datetime": "2023-02-01 17:38:45.389426"}]}\'\n)\ninstance = BaseC.from_json(payload=payload)\n\nprint(instance)\n```\n```bash\nBaseC(a_base_a=BaseA(a_datetime=datetime.datetime(2023, 2, 1, 17, 38, 45, 389426)), a_base_b=BaseB(a_uuid=UUID(\'1d9cf695-c917-49ce-854b-4063f0cda2e7\')), a_list_of_base_a=[BaseA(a_datetime=datetime.datetime(2023, 2, 1, 17, 38, 45, 389426))])\n```',
    'author': 'Marco Sievers de Almeida Ximit Gaia',
    'author_email': 'im.ximit@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""
Know your object a __init__ type validator
"""

import inspect
from typing import Union, Optional, Coroutine, Callable, Dict, Type


class Kobject:
    """
    Kobject provides a __init__ attribute type checker.
    Will rise a TypeError exception with all validation errors
    Obs: dict values are not allowed
    """

    __custom_exception: Type[Exception] = None

    @classmethod
    def set_custom_exception(cls, exception: Type[Exception]):
        """
        Will change de default validation error (TypeError)
        """
        cls.__custom_exception = exception

    def __get_attribute_metadata(self) -> dict:
        _default_values = dict(inspect.signature(self.__init__).parameters.items())
        return _default_values

    def __get_sub_attribute_type(self, attr: str):
        attr_metadata = self.__get_attribute_metadata().get(attr)
        attr_type = attr_metadata.annotation.__args__
        return attr_type

    def __get_attribute_type(self, attr: str):
        attr_metadata = self.__get_attribute_metadata().get(attr)
        attr_type = attr_metadata.annotation
        is_a_wrap_type = hasattr(attr_type, "__origin__")
        if is_a_wrap_type:
            attr_type = attr_type.__origin__
        is_not_a_real_type = attr_type.__class__ in (
            Union.__class__,
            Optional.__class__,
        )
        if is_not_a_real_type:
            attr_type = self.__get_sub_attribute_type(attr)

        return attr_type

    def __attribute_has_default_value(self, attr: str, attr_value: any) -> bool:
        attr_metadata = self.__get_attribute_metadata()
        default_value = attr_metadata.get(attr).default
        return attr_value == default_value

    def __post_init__(self):
        _errors = {}
        for attr in self.__get_attribute_metadata().keys():
            attr_value = object.__getattribute__(self, attr)
            attr_type = self.__get_attribute_type(attr=attr)
            self.__check_and_update_error_list(
                attr_value=attr_value, attr_type=attr_type, _errors=_errors, attr=attr
            )
            need_valida_content_type = isinstance(attr_value, (list, tuple))
            if need_valida_content_type:
                for index, attr_value_item in enumerate(attr_value):
                    content_types = self.__get_sub_attribute_type(attr=attr)
                    self.__check_and_update_error_list(
                        attr_value=attr_value_item,
                        attr_type=content_types,
                        _errors=_errors,
                        attr=attr,
                        index=index,
                    )
        need_raise_error = bool(_errors)
        if need_raise_error:
            message = self.__format_errors(_errors=_errors)
            exception = self.__custom_exception
            if self.__custom_exception is None:
                exception = TypeError
            raise exception(message)

    @staticmethod
    def __format_errors(_errors: dict) -> str:
        _errors_string = ["Validation Errors:\n"]
        for name, errors in _errors.items():
            _errors_string += [f"    '{name}' : {error}\n" for error in errors]
        return "".join(_errors_string)

    @staticmethod
    def __save_errors(_errors: dict, message: str, attr: any):
        if attr not in _errors:
            _errors.update({attr: []})
        _errors[attr].append(message)

    def __check_and_update_error_list(
        self,
        attr_value: any,
        attr_type: any,
        _errors: dict,
        attr: any,
        index: int = None,
    ):
        is_a_dict = attr_type in (dict, Dict)
        has_dict_on_types = isinstance(attr_type, (tuple, list)) and (
            dict in attr_type or Dict in attr_type
        )
        if is_a_dict or has_dict_on_types:
            message = f"Invalid type! The attribute '{attr}' is a {dict}"
            self.__save_errors(_errors=_errors, message=message, attr=attr)
            return

        if self.__attribute_has_default_value(attr=attr, attr_value=attr_value):
            return

        expected_callback = attr_type == Callable and not inspect.isfunction(attr_value)
        expected_coroutine = attr_type == Coroutine and not inspect.iscoroutine(
            attr_value
        )
        expected_type = not isinstance(attr_value, attr_type)
        has_type_error = expected_callback or expected_coroutine or expected_type
        if has_type_error:
            message = f"Wrong type! Expected {attr_type} but giving {type(attr_value)}"
            if index is not None:
                message += f" on index {index}"
            self.__save_errors(_errors=_errors, message=message, attr=attr)

    def __repr__(self):
        class_name = self.__class__.__name__
        values = []
        for attr in self.__get_attribute_metadata().keys():
            attr_value = object.__getattribute__(self, attr)
            values.append(f"{attr}={attr_value}")
        return f"<{class_name} ({', '.join(values)})>"
